#!/usr/bin/env python3
#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import os
from pathlib import Path


class Config:
    def __init__(self) -> None:
        conf_file = self._get_config_filepath()
        self.conf_file = conf_file

    @staticmethod
    def _get_config_filepath() -> str:
        """Get the path of the config file.
        
        Returns:  The config file path.
        """
        odl_conf = os.path.join(os.path.expanduser("~"), ".opendatalab", "config.json")
        conf_dir = Path(odl_conf).parent
        if not Path(conf_dir).exists():
            conf_dir.mkdir(parents=True)

        if not Path(odl_conf).exists():
            Path(odl_conf).touch(exist_ok=True)

        return odl_conf


config = Config()



#!/usr/bin/env python3
#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import os
import sys
from pathlib import Path
from typing import List

import click
from tqdm import tqdm

from opendatalab.cli.policy import private_policy_url, service_agreement_url
from opendatalab.cli.utility import ContextInfo, exception_handler
from opendatalab.client import downloader
from opendatalab.exception import OdlDataNotExistsError

STATUS_DICT = {
    # "noAuthRequired": "No Authorization needed",
    "pendingRequirement": "In order to download this dataset, please fill in an application form via our website.",
    "waiting": "Authorization submitted, please wait for the application result",
    "rejected": "Authorization submitted, but rejected. Please contact us for more information",
    # "accepted": "Authorization submitted, download available."
}

def handler(dwCtrlType):
    if dwCtrlType == 0:  # CTRL_C_EVENT
        pid = os.getpid()
        os.kill(pid, 9)


if sys.platform == "win32":
    import win32api
    win32api.SetConsoleCtrlHandler(handler, True)
    
    
@exception_handler
def implement_get(obj: ContextInfo, name: str, destination:str, num_workers:int):
    """
    implementation for getting dataset files
    Args:
        obj (ContextInfo):
        name (str):
        thread (int):
        compressed (bool):
    Returns:
    """
    # process dataset_name and split
    # print(name)
    ds_split = name.split("/")
    dataset_name = ds_split[0]
    if ds_split[-1] == '':
        ds_split.pop()
    single_file_flag = False
    sub_dir = "/".join(ds_split[1:])
    
    if len(ds_split) >= 2 and ('.' in ds_split[-1]):
        single_file_flag = True
        if len(ds_split) == 2:
            # indicate README.md
            file_name = ds_split[-1]
            sub_dir = ''
        elif len(ds_split) > 2:
            sub_dir = "/".join(ds_split[1:-1])
            file_name = sub_dir + '/' + ds_split[-1]
                # client init
    client = obj.get_client()
    data_info = client.get_api().get_info(dataset_name)
    
    # basic info of dataset
    info_dataset_name = data_info['name']
    info_dataset_id = data_info['id']
    
    # check the download authorization status.
    auth_status = client.get_api().get_auth_status(dataset_id=info_dataset_id)
    if auth_status['state'] in STATUS_DICT.keys():
        click.echo(f"{STATUS_DICT[auth_status['state']]}")
        sys.exit(1)
        
    # get risk level
    info_dataset_risk = data_info['attrs'].get('riskLevel', 0)
    info_dataset_url = data_info['attrs']['publishUrl']
    if info_dataset_risk > 3:
        click.echo(f"Direct download for {dataset_name} is currently not available."
                   f"\nPlease visit the homepage {info_dataset_url} for more information.")
        sys.exit(1)

    
    dataset_res_dict = client.get_api().get_dataset_files(dataset_name=info_dataset_name,
                                                          prefix = sub_dir)
    if not single_file_flag:
        total_object = dataset_res_dict['total']
        # obj list constuct
        obj_info_list = []
        for info in dataset_res_dict['list']:
            curr_dict = {}
            if not info['isDir']:
                curr_dict['size'] = info['size']
                if single_file_flag:
                    curr_dict['name'] = info['path']
                elif len(sub_dir.split('/')) > 1:
                    # curr_dict['name'] = sub_dir
                    curr_dict['name'] = info['path']
                else:
                    curr_dict['name'] = info['path']
                obj_info_list.append(curr_dict)
    else:
        total_object = 1
        obj_info_list = []
        for info in dataset_res_dict['list']:
            curr_dict = {}
            if info['path'] == str(file_name):
                curr_dict['size'] = info['size']
                curr_dict['name'] = info['path']
                obj_info_list.append(curr_dict)

    local_dir = destination
    
    download_data = client.get_api().get_download_record(info_dataset_name)
    has_download = download_data['hasDownload']

    if not has_download:
        if click.confirm(f"<<User Service Agreement>>: {service_agreement_url}"
                         f"\n<<Privacy Policy>>: {private_policy_url}"
                         f"\n[Warning]: Before downloading, please agree above content."):
            client.get_api().submit_download_record(info_dataset_name, download_data)
        else:
            click.secho('See you next time~!')
            sys.exit(1)
    
    if click.confirm(f"Download files into local directory: {local_dir} ?", default=True):
        if not Path(local_dir).exists():
            Path(local_dir).mkdir(parents=True)
            print(f"create local dir: {local_dir}")
    else:
        click.secho('See you next time~!')
        sys.exit(1)


    with tqdm(total = total_object) as pbar:
        for idx in range(total_object):
            dataset_seg_list = []
            dataset_seg_list.append(obj_info_list[idx])
            download_urls_list = client.get_api().get_dataset_download_urls(
                                                            dataset_id=info_dataset_id, 
                                                            dataset_list=dataset_seg_list)
            # print(download_urls_list)
            url_download = download_urls_list[0]['url']
            filename = download_urls_list[0]['name']
            # print(url_download, filename)
            click.echo(f"Downloading No.{idx+1} of total {total_object} files")
            if os.path.exists((os.path.join(destination, info_dataset_name, filename))):
                # print(os.path.join(destination, info_dataset_name, filename))
                click.echo('target already exists, jumping to next!')
                pbar.update(1)
                continue
            
            downloader.Downloader(url = url_download, 
                                  filename= filename, 
                                  download_dir = os.path.join(destination, info_dataset_name), 
                                  blocks_num= num_workers).start()
            pbar.update(1)
            
    click.echo(f"\nDownload Complete!")


#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import re
import time

from rich import box
from rich.console import Console
from rich.table import Table

from opendatalab.cli.utility import ContextInfo, exception_handler
from opendatalab.utils import bytes2human


def rich_content_str(keywords: str, content: str):
    keywords = keywords.strip()
    len_keywords = len(keywords)
    hl_prefix = '[bold magenta]'
    hl_suffix = '[/bold magenta]'
    start_index_list = [m.start() for m in re.finditer(keywords.lower(), content.lower())]
    result = []
    curs = 0
    if start_index_list and len(start_index_list) > 0:
        for i in start_index_list:
            result.append(content[curs:i])
            result.append(hl_prefix)
            result.append(content[i: i + len_keywords])
            result.append(hl_suffix)
            curs = i + len_keywords

        if curs < len(content):
            result.append(content[curs:])

    rich_result = ''.join(result)
    if not rich_result:
        rich_result = content

    return rich_result


@exception_handler
def implement_search(obj: ContextInfo, keywords: str) -> None:
    """
    implement dataset search through backend api
    Args:
        obj (ContextInfo):
        keywords (str):

    Returns:

    """
    client = obj.get_client()
    odl_api = client.get_api()
    import time
    time_start =time.time()
    result_list = odl_api.search_dataset(keywords)
    time_end = time.time()
    # print('-------------time_consuming--------', time_end - time_start, 's')
    console = Console()
    table = Table(show_header=True, header_style='bold cyan', box=box.ASCII2)
    table.add_column("Name", min_width=10, justify='left', overflow='fold')
    table.add_column("DataType", min_width=8, justify='left', overflow='fold')
    table.add_column("FileByte", min_width=8, overflow='fold')
    table.add_column("FileCount", min_width=8, overflow='fold')
    table.add_column("TaskType", min_width=8, justify='left', overflow='fold')
    table.add_column("LabelType", min_width=8, justify='left', overflow='fold')
    table.add_column("ViewCount", min_width=8, justify='left', overflow='fold')
    table.add_column("Introduction", min_width=20, justify='left', overflow='fold')

    # TODO:
    if result_list:
        for _, res in enumerate(result_list):
            ds_name = res['name']
            ds_name_rich = rich_content_str(keywords=keywords, content=ds_name)
            ds_view_count = res['viewCount']
            ds_desc = res['introduction']['en'][:97] + '...'
            ds_desc_rich = rich_content_str(keywords=keywords, content=ds_desc)

            ds_attr_info = res['attrs']
            ds_file_byte = bytes2human(ds_attr_info.get('fileBytes', 0))
            ds_file_count = ds_attr_info.get('fileCount',0)

            ds_data_types = _get_complex_types_str(ds_attr_info, 'mediaTypes')
            ds_task_types = _get_complex_types_str(ds_attr_info, 'taskTypes')
            ds_label_types = _get_complex_types_str(ds_attr_info, 'labelTypes')

            ds_task_types_rich = rich_content_str(keywords=keywords, content=ds_task_types)
            ds_label_types_rich = rich_content_str(keywords=keywords, content=ds_label_types)

            table.add_row(ds_name_rich, ds_data_types, str(ds_file_byte), str(ds_file_count), ds_task_types_rich,
                          ds_label_types_rich, str(ds_view_count), ds_desc_rich, end_section=True)

    console.print(table)


def _get_complex_types_str(ds_attr_info, type_name):
    if not (ds_attr_info or type_name):
        return ""

    if type_name in ds_attr_info.keys():
        type_list = ds_attr_info[type_name]
        return ','.join([d['name']['en'] for d in type_list])
    else:
        return ""



#!/usr/bin/env python3
#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import http
import sys

import click
import requests
from requests.adapters import HTTPAdapter

from opendatalab.client.api import OpenDataLabAPI
from opendatalab.exception import OpenDataLabError
from opendatalab.utils import get_api_token_from_env, parse_url


class Dataset:
    def __init__(self, url: str, token: str = "", odl_cookie: str = "") -> None:
        host, dataset_name = parse_url(url)
        self.dataset_name = dataset_name
        if token == "":
            token = get_api_token_from_env()
        self.open_data_lab_api = OpenDataLabAPI(host, token, odl_cookie)

        self.oss_bucket = None
        self.bucket_name = None
        self.oss_path_prefix = ""


#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import sys

from rich import box
from rich import print as rprint
from rich.console import Console
from rich.table import Table

from opendatalab.cli.utility import ContextInfo, exception_handler
from opendatalab.exception import OdlAccessDeniedError
from opendatalab.utils import bytes2human


@exception_handler
def implement_ls(obj: ContextInfo, dataset: str):
    """
    implementation for show dataset files
    Args:
        obj (ContextInfo):
        dataset (str): dataset name

    Returns:

    """
    ds_split = dataset.split("/")
    if len(ds_split) > 1:
        dataset_name = ds_split[0]
        sub_dir = "/".join(ds_split[1:])
    else:
        dataset_name = dataset
        sub_dir = ""

    client = obj.get_client()
    info_dataset_name = client.get_api().get_info(dataset_name)['name']
    dataset_instance = client.get_dataset(dataset_name=info_dataset_name)

    dataset_res_dict = client.get_api().get_dataset_files(dataset_name=info_dataset_name, prefix = sub_dir)
    
    # generate output info dict
    object_info_dict = {}
    total_files, total_size = 0, 0
    total_files = dataset_res_dict['total']
    for info in dataset_res_dict['list']:
        object_info_dict[info['path']] = bytes2human(info['size'])
        total_size += info['size']


    if len(object_info_dict) == 0:
        raise OdlAccessDeniedError()
        sys.exit(-1)

    sorted_object_info_dict = dict(sorted(object_info_dict.items(), key=lambda x: x[0], reverse=True))

    console = Console()
    table = Table(show_header=True, header_style='bold cyan', box=box.ASCII2)  # ROUNDED
    table.add_column("File Name", min_width=20, justify='left')
    table.add_column("Size", width=12, justify='left')

    print(f"Total file count: {total_files}, Size: {bytes2human(total_size)}")
    for key, val in sorted_object_info_dict.items():
        table.add_row(key, val, end_section=True)
    console.print(table)



#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
import sys

from opendatalab.cli.utility import ContextInfo, exception_handler


@exception_handler
def implement_login(obj: ContextInfo, username: str, password: str) -> None:
    try:
        client = obj.get_client()
        odl_api = client.get_api()
        config_json = odl_api.odl_auth(account=username, password=password)
        obj.update_config(config_json)

    except Exception as e:
        print(f"Login failure with account {username}")
        sys.exit(-1)

    print(f"Login successfully as {username}")



#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#

"""Definitions of OpenDataLab Command-line Interface commands."""
import sys
from functools import partial

import click

from opendatalab.__version__ import __svc__, __url__, __version__
from opendatalab.cli.custom import CustomCommand
from opendatalab.cli.utility import ContextInfo


@click.group(context_settings={"help_option_names": ("-h", "--help")})
@click.version_option(__version__)
@click.option("-u", "--url", type=str, default=__url__, help="The login url.", hidden=True)
@click.option("-t", "--token", type=str, default="", help="odl api token", hidden=True,
              envvar="OPENDATALAB-API-TOKEN", )
@click.pass_context
def cli(ctx: click.Context, url: str, token: str) -> None:
    """You can use `odl <command>` to access open datasets.\f

    Arguments:
        token:
        ctx: The context to be passed as the first argument.
        url: The login URL.
    """
    # ctx.ensure_object(dict)
    from opendatalab.cli.utility import implement_cli
    implement_cli(ctx, url, token)
    upgrade(ctx.obj)


command = partial(cli.command, cls=CustomCommand)


@command(synopsis=("$ odl version     # show opendatalab version",))
def version():
    """Show opendatalab version.
    """
    click.echo(f"odl version, current: {__version__}, svc: {__svc__}")


# @command(synopsis=("$ odl upgrade      # check opendatalab version upgrade.",))
# @click.pass_obj
def upgrade(obj: ContextInfo):
    """upgrade opendatalab version.
    """
    from opendatalab.cli.upgrade import implement_upgrade
    implement_upgrade(obj)


@command(synopsis=("$ odl logout      # logout current account",))
@click.pass_obj
def logout(obj: ContextInfo):
    """Logout opendatalab account.\f
    Args:
        obj (ContextInfo): context info
    """
    from opendatalab.cli.logout import implement_logout
    implement_logout(obj)


@command(synopsis=("$ odl login -u -p      # login opendatalab",))
@click.option(
    "-u",
    "--username",
    prompt="Username",
    default="",
    help="Username for opendatalab",
)
@click.option(
    "-p",
    "--password",
    prompt="Password",
    hide_input=True,
    default="", help="Password for opendatalab"
)
@click.pass_obj
def login(obj: ContextInfo, username: str, password: str):
    """Login opendatalab with account.\f

    Args:
        obj (ContextInfo): context info
        username (str): account username
        password (str): account password
    """
    from opendatalab.cli.login import implement_login
    implement_login(obj, username, password)


@command(synopsis=("$ odl ls dataset              # list dataset files",))
@click.argument("name", nargs=1)
@click.pass_obj
def ls(obj: ContextInfo, name: str) -> None:
    """List files of the dataset.\f

    Args:
        obj (ContextInfo): context info
        name (str): dataset name
    """
    from opendatalab.cli.ls import implement_ls
    implement_ls(obj, name)


@command(synopsis=("$ odl search keywords      # search dataset with keywords",))
@click.argument("keywords", nargs=1)
@click.pass_obj
def search(obj: ContextInfo, keywords):
    """Search dataset info.\f

    Args:
        obj (ContextInfo): context info
        keywords (str): dataset keywords
    """
    from opendatalab.cli.search import implement_search
    implement_search(obj, keywords)


@command(synopsis=("$ odl info dataset_name      # show dataset info",))
@click.argument("name", nargs=1)
@click.pass_obj
def info(obj: ContextInfo, name):
    """Print dataset info.\f
    Args:
        obj (ContextInfo): context info
        name (str): dataset name
    """
    from opendatalab.cli.info import implement_info
    implement_info(obj, name)


@command(synopsis=("$ odl get dataset_name      # get dataset files into local",))
@click.argument("name", nargs=1)
@click.option(
    "--dest",
    "-d",
    default='',
    help="Desired dataset store path",
    show_default=True
)
@click.option(
    "--workers",
    "-w",
    default = 8,
    help= "number of workers",
    show_default = True
)
@click.pass_obj
def get(obj: ContextInfo, name, dest, workers):
    """Get(Download) dataset files into local path.\f
    Args:
        obj (ContextInfo): context info\f
        name (str): dataset name\f
        destination(str): desired dataset store path\f
        wokers(str): number of workers\f
    """
    
    from opendatalab.cli.get import implement_get
    implement_get(obj, name, dest, workers)
if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#

"""Dataset related classes."""

from opendatalab.dataset.dataset import Dataset

__all__ = [
    "Dataset",
]



#
# Copyright 2022 Shanghai AI Lab. Licensed under MIT License.
#
from opendatalab.cli.utility import ContextInfo, exception_handler
import click


@exception_handler
def implement_logout(obj: ContextInfo) -> None:
    ##TODO: add /api/users/sync/logout
    config_content = obj.get_config_content()
    username = ""
    if 'user.email' in config_content.keys():
        username = config_content['user.email']
    
    if username.strip():
        if click.confirm('Do you want to logout?'):
            print(f"{username} logout")
            obj.clean_config()
    else:
        print("Warning: you haven't login.")



# -*- coding: utf-8 -*-
import glob
import hashlib
import os
import sys
import threading
import time

import requests


class Worker:
    def __init__(self, name: str, url: str, range_start, range_end, cache_dir, finish_callback):
        self.name = name
        self.url = url
        self.cache_filename = os.path.join(cache_dir, name + ".odl")
        self.range_start = range_start  # fixed
        self.range_end = range_end  # fixed
        self.range_curser = range_start  # curser dynamic
        self.finish_callback = finish_callback
        self.terminate_flag = False
        self.FINISH_TYPE = ""  # DONE\HELP\RETIRE 

    def __run(self):
        chunk_size = 1 * 1024
        header = {
            'Range': f'Bytes={self.range_curser}-{self.range_end}', 
        }
        req = requests.get(self.url, stream=True, headers=header)

        if 200 <= req.status_code <= 299:
            with open(self.cache_filename, "wb") as cache:
                for chunk in req.iter_content(chunk_size=chunk_size):
                    if self.terminate_flag:
                        break
                    cache.write(chunk)
                    self.range_curser += len(chunk)
        if not self.terminate_flag:
            self.FINISH_TYPE = "DONE"
        req.close()
        self.finish_callback(self)

    def start(self):
        threading.Thread(target=self.__run).start()

    def help(self):
        self.FINISH_TYPE = "HELP"
        self.terminate_flag = True

    def retire(self):
        self.FINISH_TYPE = "RETIRE"
        self.terminate_flag = True

    def __lt__(self, another):
        return self.range_start < another.range_start

    def get_progress(self):
        """progress for each worker"""
        _progress = {
            "curser": self.range_curser,
            "start": self.range_start,
            "end": self.range_end
        }
        return _progress


class Downloader:
    def __init__(self, url: str, filename:str, download_dir: str, blocks_num: int = 8):
        assert 0 <= blocks_num <= 32
        self.prefix_flag = False
        if len(filename.split('/')) == 1:
            self.filename = filename
            self.prefix = ''
        else:
            self.filename = filename.split('/')[-1]
            self.prefix_flag = True
            self.prefix = os.path.dirname(filename)
        self.url = url
        self.download_dir = download_dir
        

        # self.download_dir = os.path.join(download_dir, f".{os.sep}odl{os.sep}")
        self.blocks_num = blocks_num
        self.__bad_url_flag = False
        self.file_size = self.__get_size()
        if self.file_size <= 1:
            return
        if not self.__bad_url_flag:
        # make download dir
            if not os.path.exists(self.download_dir):
                os.makedirs(self.download_dir)
                
            # make cache dir
            if self.prefix_flag:
                self.cache_dir = os.path.join(self.download_dir,self.prefix,'.cache/')
            else:
                self.cache_dir = os.path.join(self.download_dir,'.cache/')
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
            # print(self.url, self.file_size)
            # slicing
            self.start_since = time.time()
            # worker container
            self.workers = []  
            self.LOG = self.__get_log_from_cache() 
            self.__done = threading.Event()
            self.__download_record = []
            threading.Thread(target=self.__supervise).start()
            # main
            self.__main_thread_done = threading.Event()
            # 
            readable_size = self.__get_readable_size(self.file_size)
            pathfilename = os.path.join(self.download_dir, self.prefix,self.filename)

    def __get_size(self):
        try:
            # print(self.url)
            # req = requests.head(self.url)
            # print(req.headers)
            # content_length = req.headers["Content-Length"]
            resp = requests.get(self.url,stream=True)
            content_length = resp.headers["Content-Length"]
            # print(f"-------------{content_length}--------------")
            resp.close()
            # print(req.headers)
            # print(req.headers["Content-Length"])
            return int(content_length)
        except Exception as err:
            self.__bad_url_flag = True
            self.__whistleblower(f"[Error] {err}")
            return 0

    def __get_readable_size(self, size):
        units = ["B", "KB", "MB", "GB", "TB", "PB"]
        unit_index = 0
        K = 1024.0
        while size >= K:
            size = size / K
            unit_index += 1
        return "%.1f %s" % (size, units[unit_index])

    def __get_cache_filenames(self):
        return glob.glob(f"{self.cache_dir}{self.filename}.*.odl")

    def __get_ranges_from_cache(self):
        # like ./cache/filename.1120.odl
        ranges = []
        for filename in self.__get_cache_filenames():
            size = os.path.getsize(filename)
            if size > 0:
                cache_start = int(filename.split(".")[-2])
                cache_end = cache_start + size - 1
                ranges.append((cache_start, cache_end))
        ranges.sort(key=lambda x: x[0])
        return ranges

    def __get_log_from_cache(self):
        ranges = self.__get_ranges_from_cache()  
        LOG = []
        if len(ranges) == 0:
            LOG.append((0, self.file_size - 1))
        else:
            for i, (start, end) in enumerate(ranges):
                if i == 0:
                    if start > 0:
                        LOG.append((0, start - 1))
                next_start = self.file_size if i == len(ranges) - 1 else ranges[i + 1][0]
                if end < next_start - 1:
                    LOG.append((end + 1, next_start - 1))
        return LOG

    def __increase_ranges_slice(self, ranges: list, minimum_size=1024 * 1024):
        assert len(ranges) > 0
        block_size = [end - start + 1 for start, end in ranges]
        index_of_max = block_size.index(max(block_size))
        start, end = ranges[index_of_max]
        halfsize = block_size[index_of_max] // 2
        if halfsize >= minimum_size:
            new_ranges = [x for i, x in enumerate(ranges) if i != index_of_max]
            new_ranges.append((start, start + halfsize))
            new_ranges.append((start + halfsize + 1, end))
        else:
            new_ranges = ranges
        return new_ranges

    def __ask_for_work(self, worker_num: int):
        """ask for work, return[work_range],update self.LOG"""
        assert worker_num > 0
        task = []
        LOG_num = len(self.LOG)
        # no work now, ask for new work
        if LOG_num == 0:
            self.__share_the_burdern()
            return []
        # enough work, consume
        if LOG_num >= worker_num:  
            for _ in range(worker_num):
                task.append(self.LOG.pop(0))
        # too much work
        else:
            slice_num = worker_num - LOG_num 
            task = self.LOG
            self.LOG = []
            for _ in range(slice_num):
                task = self.__increase_ranges_slice(task)
        task.sort(key=lambda x: x[0])
        return task

    def __share_the_burdern(self, minimum_size=1024 * 1024):
        """Find the heavy worker, and introduce helper"""
        max_size = 0
        max_size_name = ""
        for w in self.workers:
            p = w.get_progress()
            size = p["end"] - p["curser"] + 1
            if size > max_size:
                max_size = size
                max_size_name = w.name
        if max_size >= minimum_size:
            for w in self.workers:
                if w.name == max_size_name:
                    w.help()
                    break

    def __give_back_work(self, worker: Worker):
        """Take unfinished work"""
        progress = worker.get_progress()
        curser = progress["curser"]
        end = progress["end"]
        if curser <= end:
            self.LOG.append((curser, end))
            self.LOG.sort(key=lambda x: x[0])

    def __give_me_a_worker(self, start, end):
        worker = Worker(name=f"{self.filename}.{start}",
                          url=self.url, range_start=start, range_end=end, cache_dir=self.cache_dir,
                          finish_callback=self.__on_worker_finish,
                        )
        return worker

    def __whip(self, worker: Worker):
        """assign new job"""
        self.workers.append(worker)
        self.workers.sort()
        worker.start()

    def __on_worker_finish(self, worker: Worker):
        assert worker.FINISH_TYPE != ""
        self.workers.remove(worker)
        # need helper
        if worker.FINISH_TYPE == "HELP":
            self.__give_back_work(worker)
            self.workaholic(2)
        # job done
        elif worker.FINISH_TYPE == "DONE":
            # get one more job
            self.workaholic(1)
        elif worker.FINISH_TYPE == "RETIRE":
            self.__give_back_work(worker)
        # Job Done, Sewing!
        if self.workers == [] and self.__get_log_from_cache() == []:
            self.__sew()

    def start(self):
        # workers assembly
        for start, end in self.__ask_for_work(self.blocks_num):
            worker = self.__give_me_a_worker(start, end)
            self.__whip(worker)
        # wait till done
        self.__main_thread_done.wait()

    def stop(self):
        for w in self.workers:
            w.retire()
        while len(self.workers) != 0:
            time.sleep(0.5)
        self.LOG = self.__get_log_from_cache()

    def workaholic(self, n=1):
        """ no work no life"""
        for s, e in self.__ask_for_work(n):
            worker = self.__give_me_a_worker(s, e)
            self.__whip(worker)

    def restart(self):
        self.stop()
        # worker assembly again!
        for start, end in self.__ask_for_work(self.blocks_num):
            worker = self.__give_me_a_worker(start, end)
            self.__whip(worker)

    def __supervise(self):
        """worker and download status supervisor"""
        REFRESH_INTERVAL = 2
        # serve as a time window-length
        LAG_COUNT = 5
        WAIT_TIMES_BEFORE_RESTART = 30
        SPEED_DEGRADATION_PERCENTAGE = 0.3
        self.__download_record = []
        maxspeed = 0
        wait_times = WAIT_TIMES_BEFORE_RESTART
        while not self.__done.is_set():
            dwn_size = sum([os.path.getsize(cachefile) for cachefile in self.__get_cache_filenames()])
            self.__download_record.append({"timestamp": time.time(), "size": dwn_size})
            if len(self.__download_record) > LAG_COUNT:
                self.__download_record.pop(0)
            s = self.__download_record[-1]["size"] - self.__download_record[0]["size"]
            t = self.__download_record[-1]["timestamp"] - self.__download_record[0]["timestamp"]
            if not t == 0:
                EPSILON = 1e-5
                speed = s / t
                readable_speed = self.__get_readable_size(speed)
                # print(s,t,readable_speed)
                percentage = self.__download_record[-1]["size"] / self.file_size * 100
                status_msg = f"\r[Current File Download Info] File Progress: {percentage:.2f} % | Speed: {readable_speed}/s | Number of Workers: {len(self.workers)} | Time Elapsed: {(time.time() - self.start_since):.0f}s | ETA: {((self.file_size- dwn_size)/(speed+EPSILON)):.2f}s"
                self.__whistleblower(status_msg)
                # speed monitor
                maxspeed = max(maxspeed, speed)
                # tolerance reached
                time_over = wait_times < 0
                # not finished yet
                not_finished = not self.__done.is_set()
                
                # still running fast enough
                speed_drops_significantly = (maxspeed - speed + EPSILON) / (maxspeed + EPSILON) > SPEED_DEGRADATION_PERCENTAGE
                speed_under_threshold = speed < 1024 * 1024
                scene_1 = speed_drops_significantly and speed_under_threshold
                # running slow
                scene_2 = speed < 100 * 1024  
                if time_over and not_finished and (scene_1 or scene_2):
                    self.__whistleblower("\r[info] speed degradation, restarting...")
                    self.restart()
                    maxspeed = 0
                    wait_times = WAIT_TIMES_BEFORE_RESTART
                else:
                    wait_times -= 1
            time.sleep(REFRESH_INTERVAL)

    def __sew(self):
        self.__done.set()
        chunk_size = 10 * 1024 * 1024
        with open(f"{os.path.join(self.download_dir, self.prefix, self.filename)}", "wb") as f:
            for start, _ in self.__get_ranges_from_cache():
                cache_filename = f"{self.cache_dir}{self.filename}.{start}.odl"
                with open(cache_filename, "rb") as cache_file:
                    data = cache_file.read(chunk_size)
                    while data:
                        f.write(data)
                        f.flush()
                        data = cache_file.read(chunk_size)
        self.clear()
        self.__whistleblower("\r")
        self.__main_thread_done.set()

    def __whistleblower(self, saying: str):
        sys.stdout.write(saying)

    def md5(self):
        chunk_size = 1024 * 1024
        filename = f"{os.path.join(self.download_dir, self.prefix, self.filename)}"
        md5 = hashlib.md5()
        with open(filename, "rb") as f:
            data = f.read(chunk_size)
            while data:
                md5.update(data)
                data = f.read(chunk_size)
        return md5.hexdigest()
    
    def clear(self):
        for filename in self.__get_cache_filenames():
            os.remove(filename)


import os

import setuptools

about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(
    os.path.join(here, "opendatalab", "__version__.py"), "r", encoding="utf-8"
) as f:
    exec(f.read(), about)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    version=about["__version__"],
    project_urls={
        "Bug Tracker": "https://github.com/opendatalab/opendatalab-python-sdk/issues",
    },
)
